#define _WIN32_WINNT 0x0501
#include <windows.h>
#include <winbase.h>
#include <tlhelp32.h>
#include <ddk/ntapi.h>
#include <iphlpapi.h>
#include <psapi.h>

#include <stdio.h>

typedef unsigned long ulong;
typedef unsigned long long uvlong;

uvlong msecs = 1000;			/* sleep time in ms */
uvlong nproc;					/* number of processors */

static FILE *fp;
char *argv0 = "auxstats";

/*
 * It appears the Mingw IP_ADAPTER_ADDRESSES definition in its header is
 * old or perhaps incomplete. here we add a new one with the two extra
 * fields we need (tx and rx speed).
 */
typedef struct NEW_IP_ADAPTER_ADDRESSES {
	_ANONYMOUS_UNION union {
		ULONGLONG Alignment;
		_ANONYMOUS_STRUCT struct {
			ULONG Length;
			DWORD IfIndex;
		} DUMMYSTRUCTNAME;
	} DUMMYUNIONNAME;
	struct NEW_IP_ADAPTER_ADDRESSES *Next;
	PCHAR AdapterName;
	PIP_ADAPTER_UNICAST_ADDRESS FirstUnicastAddress;
	PIP_ADAPTER_ANYCAST_ADDRESS FirstAnycastAddress;
	PIP_ADAPTER_MULTICAST_ADDRESS FirstMulticastAddress;
	PIP_ADAPTER_DNS_SERVER_ADDRESS FirstDnsServerAddress;
	PWCHAR DnsSuffix;
	PWCHAR Description;
	PWCHAR FriendlyName;
	BYTE PhysicalAddress[MAX_ADAPTER_ADDRESS_LENGTH];
	DWORD PhysicalAddressLength;
	DWORD Flags;
	DWORD Mtu;
	DWORD IfType;
	IF_OPER_STATUS OperStatus;
	DWORD Ipv6IfIndex;
	DWORD ZoneIndices[16];
	PIP_ADAPTER_PREFIX FirstPrefix;
	ULONG64 TransmitLinkSpeed;
	ULONG64 ReceiveLinkSpeed;
} NEW_IP_ADAPTER_ADDRESSES;

static void
sysfatal(char *s, ...)
{
	char msg[1024];
	va_list args;

	fprintf(stderr, "%s: ", argv0);

	va_start(args, s);
	vfprintf(stderr, s, args);
	va_end(args);

	if(GetLastError()){
		FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL,
			GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			msg, sizeof(msg), 0);
		fprintf(stderr, " - %s", msg);
		SetLastError(0);
	}
	fputc('\n', stderr);
	fflush(stderr);
	exit(1);
}

static int
xtlb(int first)
{
	SYSTEM_LOOKASIDE_INFORMATION sli;

	if(NtQuerySystemInformation(SystemLookasideInformation, &sli, sizeof(sli), NULL) != NO_ERROR)
		return 0;
	fprintf(fp, "tlbmiss	=%lld 1000\n", (uvlong)sli.AllocateMisses);
	fprintf(fp, "tlbflush	=%lld 1000\n", (uvlong)sli.TotalFrees);

}


static int
xcpuperf(int first)
{
	int tries, err;
	uvlong idle, kern, user, intr;
	static DWORD len = 0;
	static SYSTEM_PROCESSOR_TIMES *inf = NULL;

	for(tries = 0; tries < 4; tries++){
		memset(inf, 0, len);
		err = NtQuerySystemInformation(SystemProcessorTimes, inf, len, &len);
		if(err != STATUS_INFO_LENGTH_MISMATCH)
			break;
		if((inf = realloc(inf, len)) == NULL)
			sysfatal("xcpuperf: No memory\n");
	}
	if(err != NO_ERROR)
		return 0;

	/*
	 * DPCs (Deferred Procedure Calls) appear to be upper half driver code,
	 * and so I think it makes sense to charge their time to the kernel rather
	 * than interrupt handlers.
	 *
 	 * Kernel time appears to include idle time.
	 */

	idle = inf->IdleTime.QuadPart;
	kern = inf->KernelTime.QuadPart + inf->DpcTime.QuadPart - idle;
	user = inf->UserTime.QuadPart;
	intr = inf->InterruptTime.QuadPart;
	
	fprintf(fp, "idle	%lld %lld\n", idle, msecs*nproc*500);
	fprintf(fp, "user	%lld %lld\n", user, msecs*nproc*500);
	fprintf(fp, "sys	%lld %lld\n", kern, msecs*nproc*500);
	fprintf(fp, "intr	%lld %lld\n", intr, msecs*nproc*500);
	fprintf(fp, "cpu	%lld %lld\n", user+kern, msecs*nproc*500);

/* there are better ways of getting this kind of info */
	fprintf(fp, "load	%lld %lld\n", kern + user + intr, msecs*nproc*500);
	return 0;
}


static int
xsysperf(int first)
{
	int tries, err;
	static DWORD len = 0;
	static SYSTEM_PERFORMANCE_INFORMATION *inf = NULL;

	for(tries = 0; tries < 4; tries++){
		memset(inf, 0, len);
		err = NtQuerySystemInformation(SystemPerformanceInformation, inf, len, &len);
		if(err != STATUS_INFO_LENGTH_MISMATCH)
			break;
		if((inf = realloc(inf, len)) == NULL)
			sysfatal("xsysperf: No memory\n");
	}
	if(err != NO_ERROR)
		return 0;

	fprintf(fp, "fault		%lld %lld\n", (uvlong)inf->PageFaults, msecs*nproc*20);
	fprintf(fp, "context	%lld %lld\n", (uvlong)inf->ContextSwitches, msecs*nproc*40);
	fprintf(fp, "syscall	%lld %lld\n", (uvlong)inf->SystemCalls, msecs*nproc*400);
	return 0;
}

static int
xperfinfo(int first)
{
	PERFORMANCE_INFORMATION pi;

	if(GetPerformanceInfo(&pi, sizeof(pi)) == 0)
		return 0;

	fprintf(fp, "fork	%ld %ld\n", pi.ThreadCount, nproc*10);
	fprintf(fp, "swap	=%ld %ld\n", pi.CommitTotal, pi.CommitLimit);		/* not swap but pagefile use */
	fprintf(fp, "mem	=%ld %ld\n", pi.PhysicalTotal - pi.PhysicalAvailable, pi.PhysicalTotal);
	return 0;
}

static uvlong
netspeed(void)
{
	int tries, err;
	uvlong s, totspeed;
	NEW_IP_ADAPTER_ADDRESSES *aa;

	static ULONG len = 0;
	static void *buf = NULL;

	totspeed = 0;
	len = sizeof(NEW_IP_ADAPTER_ADDRESSES);
	if((buf = realloc(buf, len)) == NULL)
		sysfatal("No memory\n");

	for(tries = 0; tries < 4; tries++){
		memset(buf, 0, len);
		err = GetAdaptersAddresses(AF_UNSPEC, 0, NULL, buf, &len);
		if(err != ERROR_BUFFER_OVERFLOW)
			break;
		if((buf = realloc(buf, len)) == NULL)
			sysfatal("No memory\n");
	}
	if(err != NO_ERROR)
		return 0;

	for(aa = buf; aa; aa = aa->Next){
 		if(aa->OperStatus != IfOperStatusUp)
			continue;
		if(aa->IfType == IF_TYPE_SOFTWARE_LOOPBACK)
			continue;
		totspeed += (uvlong)(aa->TransmitLinkSpeed + aa->ReceiveLinkSpeed) / 200000LL;
	}
	return totspeed;
}

static int
xnetwork(int first)
{
    int i, err, tries;
	MIB_IFROW iface;
	uvlong totb, totp, tote, totin, totou, totinb, totoub, totmtu;

	static uvlong totkbps = 0;
	static DWORD len = 0;
    static MIB_IPADDRTABLE *iptab = NULL;

	len = 128;
	if((iptab = malloc(len)) == NULL)
		sysfatal("xnetwork1: No memory\n");
	for(tries = 0; tries < 4; tries++){
		memset(iptab, 0, len);
		err = GetIpAddrTable(iptab, &len, 0);
		if(err != ERROR_BUFFER_OVERFLOW)
			break;
		if((iptab = realloc(iptab, len)) == NULL)
			sysfatal("xnetwork2: No memory %d\n", len);
	}
	if(err != NO_ERROR)
		return 0;

 	totb = 0;
	tote = 0;
	totp = 0;
	totin = 0;
	totou = 0;
	totinb = 0;
	totoub = 0;
	totmtu = 0;
	totkbps = 0;
	for (i = 0; i < iptab->dwNumEntries; i++){
		memset(&iface, 0, sizeof(iface));
		iface.dwIndex = iptab->table[i].dwIndex;
		if(GetIfEntry(&iface) != NO_ERROR)
			continue;

		totin += iface.dwInUcastPkts + iface.dwInNUcastPkts;
		totou += iface.dwOutUcastPkts + iface.dwOutNUcastPkts;
		totp += iface.dwInUcastPkts + iface.dwInNUcastPkts +
				iface.dwOutUcastPkts + iface.dwOutNUcastPkts;
		totinb += iface.dwInOctets;
		totoub += iface.dwOutOctets;
		totb += iface.dwInOctets + iface.dwOutOctets;
		tote += iface.dwInErrors + iface.dwInDiscards +
				iface.dwInUnknownProtos + iface.dwOutErrors +
				iface.dwOutDiscards;
		totmtu += iface.dwMtu;
	}
	free(iptab);

	/*
	 * we should be able to accumulate iface.dwSpeed 
	 * but the data in it is unreliable.
	 */
	totkbps = netspeed();

	fprintf(fp, "ether %llu %llu\n", totp, totkbps*2);
	fprintf(fp, "etherin %llu %llu\n", totin, totkbps);
	fprintf(fp, "etherout %llu %llu\n", totou, totkbps);

	fprintf(fp, "etherb %llu %llu\n", totb, totkbps*totmtu*2);
	fprintf(fp, "etherinb %llu %llu\n", totinb, totkbps*totmtu);
	fprintf(fp, "etheroutb %llu %llu\n", totoub, totkbps*totmtu);

	fprintf(fp, "ethererr %llu %d\n", tote, totkbps/100);
	return 0;
}

static char *
winver(SYSTEM_INFO *si, OSVERSIONINFOEX *ver)
{
	char *s;
	int id, type;
	static char buf[32];

	id = ver->dwPlatformId;
	type = ver->wProductType;
	switch((ver->dwMajorVersion << 8) | ver->dwMinorVersion){
	case 0x0400:
    	if(id == VER_PLATFORM_WIN32_WINDOWS)
			return "95";
    	if(id == VER_PLATFORM_WIN32_NT)
			return "NT 4.0";
		break;
	case 0x040a:
 		if(id == VER_PLATFORM_WIN32_WINDOWS)
			return "98";
		break;
	case 0x045a:
 		if(id == VER_PLATFORM_WIN32_WINDOWS)
			return "ME";
	case 0x0333:
		return "NT 3.51";
		break;
	case 0x0500:
		return "2k";
		break;
	case 0x0501:
		return "XP";
		break;
	case 0x0502:
		if(type == VER_NT_WORKSTATION &&
		   si->dwProcessorType == PROCESSOR_ARCHITECTURE_AMD64)
				return "XP Pro64";
			else
				return "2k3";
		break;
	case 0x0600:
		if(type == VER_NT_WORKSTATION)
			return "Vista";
		else
			return "2008";
		break;
	case 0x0601:
		if(type == VER_NT_WORKSTATION)
			return "7";
		else
			return "2008 r2";
		break;
	case 0x0602:
		if(type == VER_NT_WORKSTATION)
			return "8";
		else
			return "2012";
		break;
	case 0x0603:
		if(type == VER_NT_WORKSTATION)
			return "8.1";
		else
			return "2012 r2";
		break;
	default:
		break;
	}
	if(type == VER_NT_WORKSTATION)
		s = "NT";
 	if(id == VER_PLATFORM_WIN32_WINDOWS)
		s = "w32";
	snprintf(buf, sizeof(buf), "%s %ld.%ld", s,
		ver->dwMajorVersion, ver->dwMinorVersion);
	return buf;
}

static int
xwindows(int first)
{
	char *plat;
	SYSTEM_INFO si;
	OSVERSIONINFOEX ver;
	HKEY hk;
	ulong mhz, len;

	if(first){
		if(RegOpenKeyEx(HKEY_LOCAL_MACHINE,
			"HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0",
			0, KEY_READ, &hk) == 0){

			len = sizeof(mhz);
			if(RegQueryValueEx(hk, "~MHz", NULL, NULL, (void *)&mhz, &len) != 0)
				mhz = 0;
			RegCloseKey(hk);
		}
		
		GetSystemInfo(&si);
		nproc = si.dwNumberOfProcessors;
		switch(si.wProcessorArchitecture){
			case PROCESSOR_ARCHITECTURE_INTEL: plat = "x86"; break;
			case PROCESSOR_ARCHITECTURE_ARM: plat = "arm"; break;
			case PROCESSOR_ARCHITECTURE_IA64: plat = "ia64"; break;
			case PROCESSOR_ARCHITECTURE_AMD64: plat = "amd64"; break;
			default: plat = "unknown"; break;
		}

		ver.dwOSVersionInfoSize = sizeof(ver);
		GetVersionEx((OSVERSIONINFO *)&ver);

		fp = fdopen(1, "w");
		setvbuf(fp, NULL, _IOFBF, 8192);

		fprintf(fp, "info	'Win %s, %s %.1fGhz' %d\n",
			winver(&si, &ver), plat, ((double)mhz / 1000.0) + 0.05, nproc);
		return 0;
	}
	fflush(fp);
	if(ferror(fp))
		return -1;
	Sleep(msecs);
	return 0;
}


int (*statfn[])(int) =
{
	xwindows,
	xcpuperf,
	xsysperf,
	xperfinfo,
	xnetwork,
	xtlb,
	NULL
};

